#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import functools
import json
import random
import time
from typing import (
    Callable,
    Literal,
    Optional,
    ParamSpec,
    TypeVar,
)

import httpx
import neptune_api.errors
from neptune_api.types import Response

from ... import exceptions
from ...warnings import (
    Http5xxWarning,
    Http429Warning,
    Http503Warning,
    HttpOtherWarning,
)
from .. import env
from ..logger import get_logger
from ..warnings import throttled_warn

logger = get_logger()

T = ParamSpec("T")
R = TypeVar("R")


def handle_errors_default(func: Callable[T, Response[R]]) -> Callable[T, Response[R]]:
    return retry_backoff(
        max_tries=None,
        soft_max_time=env.NEPTUNE_QUERY_RETRY_SOFT_TIMEOUT.get(),
        hard_max_time=env.NEPTUNE_QUERY_RETRY_HARD_TIMEOUT.get(),
        backoff_strategy=exponential_backoff(jitter="full"),
    )(handle_api_errors(func))


def exponential_backoff(
    backoff_base: float = 0.5,
    backoff_factor: float = 2.0,
    backoff_max: float = 30.0,
    jitter: Optional[Literal["full", "equal"]] = None,
) -> Callable[[int], float]:
    def _calculate_sleep(tries: int) -> float:
        sleep = backoff_base * (backoff_factor ** (tries - 1))
        if jitter == "full":
            sleep *= random.uniform(0.0, 1.0)
        elif jitter == "equal":
            sleep *= random.uniform(0.5, 1.0)
        return min(sleep, backoff_max)

    return _calculate_sleep


def retry_backoff(
    max_tries: Optional[int] = None,
    soft_max_time: Optional[float] = None,
    hard_max_time: Optional[float] = None,
    backoff_strategy: Callable[[int], float] = exponential_backoff(),
) -> Callable[[Callable[T, Response[R]]], Callable[T, Response[R]]]:
    def decorator(func: Callable[T, Response[R]]) -> Callable[T, Response[R]]:
        @functools.wraps(func)
        def wrapper(*args: T.args, **kwargs: T.kwargs) -> Response[R]:
            total_tries = 0
            backoff_tries = 0
            start_time = time.monotonic()
            last_exc = None
            last_response = None
            rate_limit_time_extension = 0.0

            while True:
                response = None
                try:
                    response = func(*args, **kwargs)
                    if 200 <= response.status_code.value < 300:
                        return response
                except exceptions.NeptuneError:
                    raise  # give up immediately on NeptuneError
                except Exception as e:
                    last_exc = e

                if response is not None:
                    last_response = response

                total_tries += 1
                backoff_tries += 1
                if max_tries is not None and total_tries >= max_tries:
                    break

                if response is not None and "retry-after" in response.headers:
                    sleep_time = float(response.headers["retry-after"])
                    rate_limit_time_extension += sleep_time
                    backoff_tries = 0  # reset backoff tries counter when using a different strategy
                    logger.debug(
                        f"Neptune API request was rate limited. Retry-after header value: {sleep_time} seconds. "
                        f"Total time spent on rate limiting so far: {rate_limit_time_extension} seconds."
                    )
                    throttled_warn(
                        Http429Warning(
                            "Neptune API request was rate limited. We will slow down and retry automatically."
                        ),
                        stacklevel=4,
                    )
                elif response is not None and response.status_code.value == 503:
                    sleep_time = backoff_strategy(backoff_tries)
                    logger.debug(
                        f"Neptune API request failed. Backoff strategy recommends backing off for {sleep_time:.2f} "
                        f"seconds. Response: {response}. Last exception: {last_exc}."
                    )
                    throttled_warn(
                        Http503Warning("Neptune API is temporarily unavailable. We will retry automatically."),
                        stacklevel=4,
                    )
                elif response is not None and 500 <= response.status_code.value < 600:
                    sleep_time = backoff_strategy(backoff_tries)
                    logger.debug(
                        f"Neptune API request failed. Backoff strategy recommends backing off for {sleep_time:.2f} "
                        f"seconds. Response: {response}. Last exception: {last_exc}."
                    )
                    throttled_warn(
                        Http5xxWarning(
                            f"Neptune API request failed with HTTP code {response.status_code.value}. "
                            "We will retry automatically."
                        ),
                        stacklevel=4,
                    )
                else:
                    sleep_time = backoff_strategy(backoff_tries)
                    logger.debug(
                        f"Neptune API request failed. Backoff strategy recommends backing off for {sleep_time:.2f} "
                        f"seconds. Response: {response}. Last exception: {last_exc}."
                    )
                    throttled_warn(
                        HttpOtherWarning("Neptune API request failed. We will retry automatically"), stacklevel=4
                    )

                elapsed_time = time.monotonic() - start_time

                remaining_time = float("inf")
                if hard_max_time is not None:
                    remaining_time = min(remaining_time, hard_max_time - elapsed_time)
                if soft_max_time is not None:
                    remaining_time = min(remaining_time, soft_max_time + rate_limit_time_extension - elapsed_time)
                if remaining_time <= 0:
                    break
                sleep_time = min(remaining_time, sleep_time)
                logger.info(f"Backing off Neptune API request for {sleep_time:.2f} seconds (try {total_tries}).")
                time.sleep(sleep_time)

            # No more retries left
            elapsed_time = time.monotonic() - start_time
            if last_response:
                error = exceptions.NeptuneRetryError(
                    total_tries, elapsed_time, last_response.status_code.value, last_response.content
                )
            else:
                error = exceptions.NeptuneRetryError(total_tries, elapsed_time)
            if last_exc:
                raise error from last_exc
            else:
                raise error

        return wrapper

    return decorator


def handle_api_errors(func: Callable[T, Response[R]]) -> Callable[T, Response[R]]:
    @functools.wraps(func)
    def wrapper(*args: T.args, **kwargs: T.kwargs) -> Response[R]:
        # In general, exceptions - subtypes of NeptuneError won't be retried - raising them makes the error final.
        # Other exceptions may be retried with a backoff - if persistent, they will eventually cause NeptuneRetryError.

        try:
            response = func(*args, **kwargs)

            _raise_for_error_code(response.status_code.value, response.content)
            return response
        except neptune_api.errors.ApiKeyRejectedError as e:
            # The API token is explicitly rejected by the backend -- don't retry anymore.
            raise exceptions.NeptuneInvalidCredentialsError() from e
        except neptune_api.errors.UnableToParseResponse as e:
            # Allow invalid 5xx errors to be retried - _raise_for_error_code will filter them out
            # reraise the original error (to retry it) otherwise
            if 500 <= e.response.status_code < 600:
                raise
            else:
                raise exceptions.NeptuneUnexpectedResponseError(
                    status_code=e.response.status_code,
                    content=e.response.content,
                )
        except httpx.TimeoutException as e:
            logger.warning(
                "Neptune API request timed out. Retrying...\n"
                "Check your network connection or increase the timeout by setting the "
                f"{env.NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS.name} environment variable "
                f"(currently: {env.NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS.get()} seconds)."
            )
            raise e
        except Exception as e:
            raise e

    return wrapper


def _raise_for_error_code(status_code: int, content: bytes) -> None:
    """
    Raise an exception for the given response status code and content if it indicates a permanent error.

    2xx status codes are considered successful responses and do not raise an exception.
    429, and 5xx status codes are considered transient and will be handled by the retry mechanism.
    3xx, and other 4xx status codes are unexpected responses and will raise an exception.

    If the content is a valid JSON and contains the "errorType" key, it is checked against known exceptions.
    If it is not the case, or the response does not match any known exceptions,
    `NeptuneUnexpectedResponseError` is raised.
    """
    if 200 <= status_code < 300 or status_code == 429 or 500 <= status_code < 600:
        return

    try:
        if content is not None:
            json_content = json.loads(content.decode("utf-8"))
            error_type = json_content.get("errorType")

            if error_type == "ACCESS_DENIED":
                raise exceptions.NeptuneProjectInaccessible()
    except (UnicodeDecodeError, json.JSONDecodeError):
        pass

    raise exceptions.NeptuneUnexpectedResponseError(
        status_code=status_code,
        content=content,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools as ft
from typing import (
    Any,
    Optional,
    Union,
)

from neptune_api.api.retrieval import get_multiple_float_series_values_proto
from neptune_api.client import AuthenticatedClient
from neptune_api.models import FloatTimeSeriesValuesRequest
from neptune_api.proto.neptune_pb.api.v1.model.series_values_pb2 import ProtoFloatSeriesValuesResponseDTO

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from .. import identifiers
from ..logger import get_logger
from ..retrieval import (
    retry,
    util,
)
from .search import ContainerType

logger = get_logger()

# Tuples are used here to enhance performance
FloatPointValue = tuple[float, float, float, bool, float]
(
    TimestampIndex,
    StepIndex,
    ValueIndex,
    IsPreviewIndex,
    PreviewCompletionIndex,
) = range(5)

TOTAL_POINT_LIMIT: int = 1_000_000


def fetch_multiple_series_values(
    client: AuthenticatedClient,
    run_attribute_definitions: list[identifiers.RunAttributeDefinition],
    include_inherited: bool,
    container_type: ContainerType,
    include_preview: bool,
    step_range: tuple[Union[float, None], Union[float, None]] = (None, None),
    tail_limit: Optional[int] = None,
) -> dict[identifiers.RunAttributeDefinition, list[FloatPointValue]]:
    if not run_attribute_definitions:
        return {}

    assert len(run_attribute_definitions) <= TOTAL_POINT_LIMIT, (
        f"The number of requested attributes {len(run_attribute_definitions)} exceeds the maximum limit of "
        f"{TOTAL_POINT_LIMIT}. Please reduce the number of attributes."
    )

    width = len(str(len(run_attribute_definitions) - 1))
    request_id_to_attribute: dict[str, identifiers.RunAttributeDefinition] = {
        f"{i:0{width}d}": attr for i, attr in enumerate(run_attribute_definitions)
    }

    params: dict[str, Any] = {
        "requests": [
            {
                "requestId": request_id,
                "series": {
                    "holder": {
                        "identifier": str(run_attribute.run_identifier),
                        "type": "experiment",
                    },
                    "attribute": run_attribute.attribute_definition.name,
                    "lineage": "FULL" if include_inherited else "NONE",
                    "lineageEntityType": "EXPERIMENT" if container_type == ContainerType.EXPERIMENT else "RUN",
                    "includePreview": include_preview,
                },
            }
            for request_id, run_attribute in request_id_to_attribute.items()
        ],
        "stepRange": {"from": step_range[0], "to": step_range[1]},
        # Fetch in descending order to enable efficient tail fetching
        "order": "descending",
    }

    results: dict[identifiers.RunAttributeDefinition, list[FloatPointValue]] = {
        run_attribute: [] for run_attribute in run_attribute_definitions
    }

    for page_result in util.fetch_pages(
        client=client,
        fetch_page=_fetch_metrics_page,
        process_page=ft.partial(_process_metrics_page, request_id_to_attribute=request_id_to_attribute),
        make_new_page_params=ft.partial(
            _make_new_metrics_page_params,
            request_id_to_attribute=request_id_to_attribute,
            tail_limit=tail_limit,
            partial_results=results,
        ),
        initial_params=params,
    ):
        for attribute, values in page_result.items:
            results[attribute].extend(values)

    # Reverse the order of values to maintain ascending order
    for attribute in results:
        results[attribute].reverse()

    return results


def _fetch_metrics_page(
    client: AuthenticatedClient,
    params: dict[str, Any],
) -> ProtoFloatSeriesValuesResponseDTO:
    logger.debug(f"Calling get_multiple_float_series_values_proto with params: {params}")

    body = FloatTimeSeriesValuesRequest.from_dict(params)
    call_api = retry.handle_errors_default(
        with_neptune_client_metadata(get_multiple_float_series_values_proto.sync_detailed)
    )
    response = call_api(client=client, body=body)

    logger.debug(
        f"get_multiple_float_series_values_proto response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )
    return ProtoFloatSeriesValuesResponseDTO.FromString(response.content)


def _process_metrics_page(
    data: ProtoFloatSeriesValuesResponseDTO,
    request_id_to_attribute: dict[str, identifiers.RunAttributeDefinition],
) -> util.Page[tuple[identifiers.RunAttributeDefinition, list[FloatPointValue]]]:
    result = {}
    for series in data.series:
        run_attribute = request_id_to_attribute[series.requestId]
        result[run_attribute] = [
            (
                point.timestamp_millis,
                point.step,
                point.value,
                point.is_preview,
                point.completion_ratio,
            )
            for point in series.series.values
        ]
    return util.Page(items=list(result.items()))


def _make_new_metrics_page_params(
    params: dict[str, Any],
    data: Optional[ProtoFloatSeriesValuesResponseDTO],
    request_id_to_attribute: dict[str, identifiers.RunAttributeDefinition],
    tail_limit: Optional[int],
    partial_results: dict[identifiers.RunAttributeDefinition, list[FloatPointValue]],
) -> Optional[dict[str, Any]]:
    if data is None:  # no past data, we are fetching the first page
        for request in params["requests"]:
            if "afterStep" in request:
                del request["afterStep"]
        per_series_points_limit = max(1, TOTAL_POINT_LIMIT // len(params["requests"]))
        if tail_limit is not None:
            per_series_points_limit = min(per_series_points_limit, tail_limit)
        params["perSeriesPointsLimit"] = per_series_points_limit
        return params

    prev_per_series_points_limit = params["perSeriesPointsLimit"]

    new_request_after_steps = {}
    for series in data.series:
        request_id = series.requestId
        value_size = len(series.series.values)
        is_page_full = value_size == prev_per_series_points_limit

        attribute = request_id_to_attribute[request_id]
        need_more_points = len(partial_results[attribute]) < tail_limit if tail_limit is not None else True

        if is_page_full and need_more_points:
            new_request_after_steps[request_id] = series.series.values[-1].step

    if not new_request_after_steps:  # no data left to fetch, return None to stop
        return None

    new_requests = []
    for request in params["requests"]:
        request_id = request["requestId"]
        if request_id in new_request_after_steps:
            after_step = new_request_after_steps[request_id]
            request["afterStep"] = after_step
            new_requests.append(request)
    params["requests"] = new_requests

    per_series_points_limit = max(1, TOTAL_POINT_LIMIT // len(params["requests"]))
    if tail_limit is not None:
        already_fetched = next(
            len(partial_results[request_id_to_attribute[request_id]]) for request_id in new_request_after_steps.keys()
        )  # assumes the results for all unfinished series have the same length
        per_series_points_limit = min(per_series_points_limit, tail_limit - already_fetched)
    params["perSeriesPointsLimit"] = per_series_points_limit

    return params



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from dataclasses import dataclass
from typing import (
    Any,
    Generator,
    Optional,
    Tuple,
)

import numpy as np
import pandas as pd

from .. import types
from ..exceptions import ConflictingAttributeTypes
from . import identifiers
from .retrieval import (
    metric_buckets,
    metrics,
    series,
)
from .retrieval.attribute_types import (
    TYPE_AGGREGATIONS,
    File,
    Histogram,
)
from .retrieval.attribute_values import AttributeValue
from .retrieval.metrics import (
    IsPreviewIndex,
    PreviewCompletionIndex,
    StepIndex,
    TimestampIndex,
    ValueIndex,
)
from .retrieval.search import ContainerType

__all__ = (
    "create_runs_table",
    "create_runs_table_multiproject",
    "TableRow",
    "create_metrics_dataframe",
    "create_series_dataframe",
    "create_files_dataframe",
    "create_metric_buckets_dataframe",
)


@dataclass
class TableRow:
    values: list[AttributeValue]
    label: str
    project_identifier: Optional[identifiers.ProjectIdentifier] = None


def create_runs_table(
    table_rows: list[TableRow],
    *,
    type_suffix_in_column_names: bool,
    container_type: ContainerType,
) -> pd.DataFrame:
    return _convert_table_to_dataframe(
        table_rows=table_rows,
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=container_type,
        include_project_column=False,
    )


def create_runs_table_multiproject(
    table_rows: list[TableRow],
    *,
    type_suffix_in_column_names: bool,
    container_type: ContainerType,
) -> pd.DataFrame:
    return _convert_table_to_dataframe(
        table_rows=table_rows,
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=container_type,
        include_project_column=True,
    )


def _convert_table_to_dataframe(
    *,
    table_rows: list[TableRow],
    type_suffix_in_column_names: bool,
    container_type: ContainerType,
    include_project_column: bool,
) -> pd.DataFrame:
    index_column_name = "experiment" if container_type == ContainerType.EXPERIMENT else "run"

    if not table_rows:
        return pd.DataFrame(
            index=(
                pd.MultiIndex.from_tuples([], names=["project", index_column_name])
                if include_project_column
                else pd.Index([], name=index_column_name)
            ),
            columns=pd.Index([], name="attribute"),
        )

    def convert_row(table_row: TableRow) -> dict[str, Any]:
        row: dict[str, Any] = {}
        for value in table_row.values:
            column_name = f"{value.attribute_definition.name}:{value.attribute_definition.type}"
            attribute_type = value.attribute_definition.type
            if attribute_type in TYPE_AGGREGATIONS:
                aggregation_value = value.value
                element_value = getattr(aggregation_value, "last")
                step = getattr(aggregation_value, "last_step", None)
            else:
                element_value = value.value
                step = None

            if attribute_type == "file" or attribute_type == "file_series":
                row[column_name] = _create_output_file(
                    file=element_value,
                    run_identifier=value.run_identifier,
                    label=table_row.label,
                    container_type=container_type,
                    attribute_path=value.attribute_definition.name,
                    step=step,
                )
            elif attribute_type == "histogram" or attribute_type == "histogram_series":
                row[column_name] = _create_output_histogram(element_value)
            else:
                row[column_name] = element_value
        return row

    def transform_column_names(df: pd.DataFrame) -> pd.DataFrame:
        if type_suffix_in_column_names:
            return df

        # Transform the column by removing the type
        original_columns = df.columns
        df.columns = pd.Index([col.rsplit(":", 1)[0] for col in df.columns])

        # Check for duplicate names
        duplicated = df.columns.duplicated(keep=False)
        if duplicated.any():
            duplicated_names = df.columns[duplicated]
            duplicated_names_set = set(duplicated_names)
            conflicting_types: dict[str, set[str]] = {}
            for original_col, new_col in zip(original_columns, df.columns):
                if new_col in duplicated_names_set:
                    conflicting_types.setdefault(new_col, set()).add(original_col.rsplit(":", 1)[1])

            raise ConflictingAttributeTypes(conflicting_types.keys())  # TODO: add conflicting types to the exception

        return df

    rows = []
    for table_row in table_rows:
        if include_project_column and table_row.project_identifier is None:
            raise ValueError("Missing project identifier for at least one row.")
        row: Any = convert_row(table_row)
        row[index_column_name] = table_row.label
        if include_project_column:
            row["project"] = str(table_row.project_identifier)
        rows.append(row)

    dataframe = pd.DataFrame(rows)
    dataframe = transform_column_names(dataframe)
    if include_project_column:
        dataframe.set_index(["project", index_column_name], drop=True, inplace=True)
    else:
        dataframe.set_index(index_column_name, drop=True, inplace=True)

    dataframe.columns.name = "attribute"
    sorted_columns = sorted(dataframe.columns)
    dataframe = dataframe[sorted_columns]

    return dataframe


def create_metrics_dataframe(
    metrics_data: dict[identifiers.RunAttributeDefinition, list[metrics.FloatPointValue]],
    sys_id_label_mapping: dict[identifiers.SysId, str],
    *,
    type_suffix_in_column_names: bool,
    include_point_previews: bool,
    index_column_name: str,
    timestamp_column_name: Optional[str] = None,
) -> pd.DataFrame:
    """
    Creates a memory-efficient DataFrame directly from FloatPointValue tuples.

    Note that `data_points` must be sorted by (experiment name, path) to ensure correct
    categorical codes.

    There is an intermediate processing step where we represent paths as categorical codes
    Example:

    Assuming there are 2 user columns called "foo" and "bar", 2 steps each. The intermediate
    DF will have this shape:

            experiment  path   step  value
        0     exp-name     0    1.0    0.0
        1     exp-name     0    2.0    0.5
        1     exp-name     1    1.0    1.5
        1     exp-name     1    2.0    2.5

    `path_mapping` would contain {"foo": 0, "bar": 1}. The column names will be restored before returning the
    DF, which then can be sorted based on its columns.

    The reason for the intermediate representation is that logging a metric called eg. "step" would conflict
    with our "step" column during df.reset_index(), and we would crash.
    Operating on integer codes is safe, as they can never appear as valid metric names.

    If `timestamp_column_name` is provided, timestamp will be included in the DataFrame under the
    specified column.
    """

    path_mapping: dict[str, int] = {}
    sys_id_mapping: dict[str, int] = {}
    label_mapping: list[str] = []

    for run_attr_definition in metrics_data:
        if run_attr_definition.run_identifier.sys_id not in sys_id_mapping:
            sys_id_mapping[run_attr_definition.run_identifier.sys_id] = len(sys_id_mapping)
            label_mapping.append(sys_id_label_mapping[run_attr_definition.run_identifier.sys_id])

        if run_attr_definition.attribute_definition.name not in path_mapping:
            path_mapping[run_attr_definition.attribute_definition.name] = len(path_mapping)

    def generate_categorized_rows() -> Generator[Tuple, None, None]:
        for attribute, points in metrics_data.items():
            exp_category = sys_id_mapping[attribute.run_identifier.sys_id]
            path_category = path_mapping[attribute.attribute_definition.name]

            for point in points:
                # Only include columns that we know we need. Note that the list of columns must match the
                # the list of `types` below.
                head = (
                    exp_category,
                    path_category,
                    point[StepIndex],
                    point[ValueIndex],
                )
                if include_point_previews and timestamp_column_name:
                    tail: Tuple[Any, ...] = (
                        point[TimestampIndex],
                        point[IsPreviewIndex],
                        point[PreviewCompletionIndex],
                    )
                elif timestamp_column_name:
                    tail = (point[TimestampIndex],)
                elif include_point_previews:
                    tail = (point[IsPreviewIndex], point[PreviewCompletionIndex])
                else:
                    tail = ()

                yield head + tail

    types = [
        (index_column_name, "uint32"),
        ("path", "uint32"),
        ("step", "float64"),
        ("value", "float64"),
    ]

    if timestamp_column_name:
        types.append((timestamp_column_name, "int64"))

    if include_point_previews:
        types.append(("is_preview", "bool"))
        types.append(("preview_completion", "float64"))

    df = pd.DataFrame(
        np.fromiter(generate_categorized_rows(), dtype=types),
    )

    if timestamp_column_name:
        df[timestamp_column_name] = pd.to_datetime(df[timestamp_column_name], unit="ms", origin="unix", utc=True)

    df = _pivot_df(df, index_column_name, timestamp_column_name, extra_value_columns=types[4:])
    df = _restore_labels_in_index(df, index_column_name, label_mapping)
    df = _restore_path_column_names(df, path_mapping, "float_series" if type_suffix_in_column_names else None)
    df = _sort_index_and_columns(df, index_column_name)

    return df


def create_series_dataframe(
    series_data: dict[identifiers.RunAttributeDefinition, list[series.SeriesValue]],
    # TODO: PY-310 remove unused parameter project_identifier
    project_identifier: str,
    sys_id_label_mapping: dict[identifiers.SysId, str],
    index_column_name: str,
    timestamp_column_name: Optional[str],
) -> pd.DataFrame:
    container_type = ContainerType.EXPERIMENT if index_column_name == "experiment" else ContainerType.RUN
    experiment_mapping: dict[identifiers.SysId, int] = {}
    path_mapping: dict[str, int] = {}
    label_mapping: list[str] = []

    for run_attr_definition in series_data.keys():
        if run_attr_definition.run_identifier.sys_id not in experiment_mapping:
            experiment_mapping[run_attr_definition.run_identifier.sys_id] = len(experiment_mapping)
            label_mapping.append(sys_id_label_mapping[run_attr_definition.run_identifier.sys_id])

        if run_attr_definition.attribute_definition.name not in path_mapping:
            path_mapping[run_attr_definition.attribute_definition.name] = len(path_mapping)

    def convert_values(
        run_attribute_definition: identifiers.RunAttributeDefinition, values: list[series.SeriesValue]
    ) -> list[series.SeriesValue]:
        if run_attribute_definition.attribute_definition.type == "file_series":
            label = sys_id_label_mapping[run_attribute_definition.run_identifier.sys_id]
            return [
                series.SeriesValue(
                    step=point.step,
                    value=_create_output_file(
                        file=point.value,
                        run_identifier=run_attribute_definition.run_identifier,
                        label=label,
                        container_type=container_type,
                        attribute_path=run_attribute_definition.attribute_definition.name,
                        step=point.step,
                    ),
                    timestamp_millis=point.timestamp_millis,
                )
                for point in values
            ]
        elif run_attribute_definition.attribute_definition.type == "histogram_series":
            return [
                series.SeriesValue(
                    step=point.step,
                    value=_create_output_histogram(point.value),
                    timestamp_millis=point.timestamp_millis,
                )
                for point in values
            ]
        else:
            return values

    def generate_categorized_rows() -> Generator[Tuple, None, None]:
        for attribute, values in series_data.items():
            exp_category = experiment_mapping[attribute.run_identifier.sys_id]
            path_category = path_mapping[attribute.attribute_definition.name]
            converted_values = convert_values(attribute, values)

            if timestamp_column_name:
                for point in converted_values:
                    yield exp_category, path_category, point.step, point.value, point.timestamp_millis
            else:
                for point in converted_values:
                    yield exp_category, path_category, point.step, point.value

    types = [
        (index_column_name, "uint32"),
        ("path", "uint32"),
        ("step", "float64"),
        ("value", "object"),
    ]
    if timestamp_column_name:
        types.append((timestamp_column_name, "int64"))

    df = pd.DataFrame(
        np.fromiter(generate_categorized_rows(), dtype=types),
    )

    if timestamp_column_name:
        df[timestamp_column_name] = pd.to_datetime(df[timestamp_column_name], unit="ms", origin="unix", utc=True)

    df = _pivot_df(df, index_column_name, timestamp_column_name, extra_value_columns=types[4:])
    df = _restore_labels_in_index(df, index_column_name, label_mapping)
    df = _restore_path_column_names(df, path_mapping, None)
    df = _sort_index_and_columns(df, index_column_name)

    return df


def create_metric_buckets_dataframe(
    buckets_data: dict[identifiers.RunAttributeDefinition, list[metric_buckets.TimeseriesBucket]],
    sys_id_label_mapping: dict[identifiers.SysId, str],
    *,
    container_column_name: str,
) -> pd.DataFrame:
    """
    Output Example:

    experiment    experiment_1                                        experiment_2
    series        metrics/loss            metrics/accuracy            metrics/loss            metrics/accuracy
    bucket                   x          y                x          y            x          y               x          y
    (0.0, 20.0]       0.766337  46.899769         0.629231  29.418603     0.793347   3.618248        0.445641  16.923348
    (20.0, 40.0]     20.435899  42.001229        20.825488  11.989595    20.151307  21.244816       20.720397  20.515981
    (40.0, 60.0]     40.798869  10.429626        40.640794  10.276835    40.338434  33.692977       40.381568  15.954130
    (60.0, 80.0]     60.856616  20.633254        60.033832   0.927636    60.002655  37.048722       60.713322  49.537098
    (80.0, 100.0]    80.522183   6.084259        80.019450  39.666397    80.003379  22.569435       80.745987  42.658697
    """

    path_mapping: dict[str, int] = {}
    sys_id_mapping: dict[str, int] = {}
    label_mapping: list[str] = []

    for run_attr_definition in buckets_data:
        if run_attr_definition.run_identifier.sys_id not in sys_id_mapping:
            sys_id_mapping[run_attr_definition.run_identifier.sys_id] = len(sys_id_mapping)
            label_mapping.append(sys_id_label_mapping[run_attr_definition.run_identifier.sys_id])

        if run_attr_definition.attribute_definition.name not in path_mapping:
            path_mapping[run_attr_definition.attribute_definition.name] = len(path_mapping)

    def generate_categorized_rows() -> Generator[Tuple, None, None]:
        for attribute, buckets in buckets_data.items():
            exp_category = sys_id_mapping[attribute.run_identifier.sys_id]
            path_category = path_mapping[attribute.attribute_definition.name]

            buckets.sort(key=lambda b: (b.from_x, b.to_x))
            for ix, bucket in enumerate(buckets):
                yield (
                    exp_category,
                    path_category,
                    bucket.from_x,
                    bucket.to_x,
                    bucket.first_x if ix == 0 else bucket.last_x,
                    bucket.first_y if ix == 0 else bucket.last_y,
                )

    types = [
        (container_column_name, "uint32"),
        ("path", "uint32"),
        ("from_x", "float64"),
        ("to_x", "float64"),
        ("x", "float64"),
        ("y", "float64"),
    ]

    df = pd.DataFrame(
        np.fromiter(generate_categorized_rows(), dtype=types),
    )

    df["bucket"] = pd.IntervalIndex.from_arrays(df["from_x"], df["to_x"], closed="right")
    df = df.drop(columns=["from_x", "to_x"])

    df = df.pivot_table(
        index="bucket",
        columns=[container_column_name, "path"],
        values=["x", "y"],
        observed=True,
        dropna=True,
        sort=False,
    )

    df = _restore_labels_in_columns(df, container_column_name, label_mapping)
    df = _restore_path_column_names(df, path_mapping, None)

    # Add back any columns that were removed because they were all NaN
    if buckets_data:
        desired_columns = pd.MultiIndex.from_tuples(
            [
                (
                    dim,
                    sys_id_label_mapping[run_attr_definition.run_identifier.sys_id],
                    run_attr_definition.attribute_definition.name,
                )
                for run_attr_definition in buckets_data.keys()
                for dim in ("x", "y")
            ],
            names=["bucket", container_column_name, "metric"],
        )
        df = df.reindex(columns=desired_columns)
    else:
        # Handle empty case - create expected column structure
        df.columns = pd.MultiIndex.from_product([["x", "y"], [], []], names=["bucket", container_column_name, "metric"])

    df = df.reorder_levels([1, 2, 0], axis="columns")
    df = df.sort_index(axis="columns", level=[0, 1])
    df = df.sort_index()
    df.index.name = None
    df.columns.names = (container_column_name, "metric", "bucket")

    df = _collapse_open_buckets(df)

    return df


def _collapse_open_buckets(df: pd.DataFrame) -> pd.DataFrame:
    """
    1st returned bucket is (-inf, first_point], which we merge with the 2nd bucket (first_point, end],
    resulting in a new bucket [first_point, end].
    If there's only one bucket, it should have form (first_point, inf). We transform it to [first_point, first_point].
    """
    df.index = df.index.astype(object)  # IntervalIndex cannot mix Intervals closed from different sides

    if df.index.empty:
        return df

    if len(df.index) == 1:
        finite_value = None
        if np.isfinite(df.index[0].right) and not np.isfinite(df.index[0].left):
            finite_value = df.index[0].right
        elif np.isfinite(df.index[0].left) and not np.isfinite(df.index[0].right):
            finite_value = df.index[0].left

        if finite_value is not None:
            new_interval = pd.Interval(left=finite_value, right=finite_value, closed="both")
            df.index = pd.Index([new_interval], dtype=object)
        return df

    col_funcs = {
        "x": lambda s: s[s.first_valid_index()] if s.first_valid_index() is not None else np.nan,
        "y": lambda s: s[s.first_valid_index()] if s.first_valid_index() is not None else np.nan,
    }

    first, second = df.index[0], df.index[1]
    if first.right >= second.left - second.length * 0.5:  # floats can be imprecise, we use bucket length as a tolerance
        new_interval = pd.Interval(left=first.right, right=second.right, closed="both")
        new_row = df.iloc[0:2].apply(axis="index", func=lambda col: col_funcs[col.name[-1]](col))
        df = df.drop(index=[first, second])
        df.loc[new_interval] = new_row
        df = df.sort_index()
    else:
        new_interval = pd.Interval(left=first.right, right=first.right + second.length, closed="both")
        df.index = [new_interval] + list(df.index[1:])

    return df


def _pivot_df(
    df: pd.DataFrame,
    index_column_name: str,
    timestamp_column_name: Optional[str],
    extra_value_columns: list[tuple[str, str]],
) -> pd.DataFrame:
    # Holds all existing (experiment, step) pairs
    # This is needed because pivot_table will remove rows if they are all NaN
    observed_idx = pd.MultiIndex.from_frame(
        df[[index_column_name, "step"]]
        .astype(
            {
                index_column_name: "uint32",
                "step": "float64",
            }
        )
        .drop_duplicates()
    )

    if df.empty and timestamp_column_name:
        # Handle empty DataFrame case to avoid pandas dtype errors
        df[timestamp_column_name] = pd.Series(dtype="datetime64[ns]")

    if extra_value_columns:
        # Holds all existing columns
        # This is needed because pivot_table will remove columns if they are all NaN
        value_columns = ["value"] + [col[0] for col in extra_value_columns]
        observed_columns = pd.MultiIndex.from_tuples(
            [(value, path) for path in df["path"].unique() for value in value_columns], names=[None, "path"]
        )

        # if there are multiple value columns, don't specify them and rely on pandas to create the column multi-index
        df = df.pivot_table(
            index=[index_column_name, "step"], columns="path", aggfunc="first", observed=True, dropna=True, sort=False
        )
    else:
        observed_columns = df["path"].unique()

        # when there's only "value", define values explicitly, to make pandas generate a flat index
        df = df.pivot_table(
            index=[index_column_name, "step"],
            columns="path",
            values="value",
            aggfunc="first",
            observed=True,
            dropna=True,
            sort=False,
        )

    # Add back any columns that were removed because they were all NaN
    return df.reindex(index=observed_idx, columns=observed_columns)


def _restore_labels_in_index(
    df: pd.DataFrame,
    column_name: str,
    label_mapping: list[str],
) -> pd.DataFrame:
    if df.index.empty:
        df.index = df.index.set_levels(df.index.get_level_values(column_name).astype(str), level=column_name)
        return df

    return df.rename(index={i: label for i, label in enumerate(label_mapping)}, level=column_name)


def _restore_labels_in_columns(
    df: pd.DataFrame,
    column_name: str,
    label_mapping: list[str],
) -> pd.DataFrame:
    if df.index.empty:
        df.columns = df.columns.set_levels(df.columns.get_level_values(column_name).astype(str), level=column_name)
        return df

    return df.rename(columns={i: label for i, label in enumerate(label_mapping)}, level=column_name)


def _restore_path_column_names(
    df: pd.DataFrame, path_mapping: dict[str, int], type_suffix: Optional[str]
) -> pd.DataFrame:
    """
    Accepts an DF in an intermediate format in _create_dataframe, and the mapping of column names.
    Restores colum names in the DF based on the mapping.
    """
    # No columns to rename, simply ensure the dtype of the path column changes from categorical int to str
    if df.columns.empty:
        if isinstance(df.columns, pd.MultiIndex):
            df.columns = df.columns.set_levels(df.columns.get_level_values("path").astype(str), level="path")
        else:
            df.columns = df.columns.astype(str)
        return df

    # We need to reverse the mapping to index -> column name
    if type_suffix:
        reverse_mapping = {index: f"{path}:{type_suffix}" for path, index in path_mapping.items()}
    else:
        reverse_mapping = {index: path for path, index in path_mapping.items()}
    return df.rename(columns=reverse_mapping)


def _sort_index_and_columns(df: pd.DataFrame, index_column_name: str) -> pd.DataFrame:
    # MultiIndex DFs need to have column index order swapped: value/metric_name -> metric_name/value.
    # We also sort columns, but only after the original names have been restored.
    if isinstance(df.columns, pd.MultiIndex):
        df.columns.names = (None, None)
        df = df.swaplevel(axis="columns")
        df = df.sort_index(axis="columns", level=0, kind="stable", sort_remaining=False)
    else:
        df.columns.name = None
        df = df.sort_index(axis="columns")

    return df.sort_index(axis="index", level=[index_column_name, "step"])


def create_files_dataframe(
    file_data: dict[types.File, Optional[pathlib.Path]],
    container_type: "ContainerType",
) -> pd.DataFrame:
    index_column_name = "experiment" if container_type == ContainerType.EXPERIMENT else "run"

    if not file_data:
        return pd.DataFrame(
            index=pd.MultiIndex.from_tuples([], names=[index_column_name, "step"]),
            columns=pd.Index([], name="attribute"),
        )

    rows: list[dict[str, Any]] = []
    for file, path in file_data.items():
        row = {
            index_column_name: file.container_identifier,
            "attribute": file.attribute_path,
            "step": file.step,
            "path": str(path) if path else None,
        }
        rows.append(row)

    dataframe = pd.DataFrame(rows)
    dataframe = dataframe.pivot(index=[index_column_name, "step"], columns="attribute", values="path")

    dataframe = dataframe.sort_index()
    sorted_columns = sorted(dataframe.columns)
    return dataframe[sorted_columns]


def _create_output_file(
    *,
    file: File,
    run_identifier: identifiers.RunIdentifier,
    label: str,
    container_type: ContainerType,
    attribute_path: str,
    step: Optional[float] = None,
) -> types.File:
    return types.File(
        project_identifier=str(run_identifier.project_identifier),
        experiment_name=label if container_type == ContainerType.EXPERIMENT else None,
        run_id=label if container_type == ContainerType.RUN else None,
        attribute_path=attribute_path,
        step=step,
        path=file.path,
        size_bytes=file.size_bytes,
        mime_type=file.mime_type,
    )


def _create_output_histogram(
    histogram: Histogram,
) -> types.Histogram:
    return types.Histogram(
        type=histogram.type,
        edges=histogram.edges,
        values=histogram.values,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ("list_attributes",)

from typing import (
    Generator,
    Optional,
)

from .. import client as _client
from .. import identifiers
from ..composition import attribute_components as _components
from ..composition import (
    concurrency,
    type_inference,
)
from ..context import (
    Context,
    get_context,
    validate_context,
)
from ..filters import (
    _BaseAttributeFilter,
    _Filter,
)
from ..retrieval import (
    search,
    util,
)


def list_attributes(
    *,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    context: Optional[Context] = None,
    container_type: search.ContainerType,
) -> list[str]:
    valid_context = validate_context(context or get_context())
    client = _client.get_client(context=valid_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):
        inference_result = type_inference.infer_attribute_types_in_filter(
            client,
            project_identifier,
            filter_,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        filter_ = inference_result.get_result_or_raise()
        inference_result.emit_warnings()

        output = _components.fetch_attribute_definitions_complete(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            attribute_filter=attributes,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
            downstream=concurrency.return_value,
            container_type=container_type,
        )

        results: Generator[util.Page[identifiers.AttributeDefinition], None, None] = concurrency.gather_results(output)
        names = set()
        for page in results:
            for item in page.items:
                names.add(item.name)

        return sorted(names)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from concurrent.futures import Executor
from typing import (
    Generator,
    Literal,
    Optional,
    Tuple,
)

import pandas as pd
from neptune_api import AuthenticatedClient

from .. import identifiers
from ..client import get_client
from ..composition import attribute_components as _components
from ..composition import (
    concurrency,
    type_inference,
    validation,
)
from ..context import (
    Context,
    get_context,
    validate_context,
)
from ..filters import (
    _BaseAttributeFilter,
    _Filter,
)
from ..identifiers import ProjectIdentifier
from ..output_format import create_series_dataframe
from ..retrieval import (
    search,
    series,
    split,
)
from ..retrieval.search import ContainerType
from ..retrieval.series import SeriesValue

__all__ = ("fetch_series",)


def fetch_series(
    *,
    project_identifier: ProjectIdentifier,
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    include_time: Optional[Literal["absolute"]],
    step_range: Tuple[Optional[float], Optional[float]],
    lineage_to_the_root: bool,
    tail_limit: Optional[int],
    context: Optional[Context] = None,
    container_type: ContainerType,
) -> pd.DataFrame:
    validation.validate_step_range(step_range)
    validation.validate_tail_limit(tail_limit)
    validation.validate_include_time(include_time)
    restricted_attributes = validation.restrict_attribute_filter_type(
        attributes, type_in={"string_series", "histogram_series", "file_series"}
    )

    valid_context = validate_context(context or get_context())
    client = get_client(context=valid_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):
        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        inferred_filter = inference_result.get_result_or_raise()
        inference_result.emit_warnings()

        series_data, sys_id_to_label_mapping = _fetch_series(
            filter_=inferred_filter,
            attributes=restricted_attributes,
            client=client,
            project_identifier=project_identifier,
            step_range=step_range,
            lineage_to_the_root=lineage_to_the_root,
            tail_limit=tail_limit,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
            container_type=container_type,
        )

        df = create_series_dataframe(
            series_data,
            project_identifier,
            sys_id_to_label_mapping,
            index_column_name="experiment" if container_type == ContainerType.EXPERIMENT else "run",
            timestamp_column_name="absolute_time" if include_time == "absolute" else None,
        )

    return df


def _fetch_series(
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    step_range: tuple[Optional[float], Optional[float]],
    lineage_to_the_root: bool,
    tail_limit: Optional[int],
    container_type: ContainerType,
) -> tuple[dict[identifiers.RunAttributeDefinition, list[SeriesValue]], dict[identifiers.SysId, str]]:
    sys_id_label_mapping: dict[identifiers.SysId, str] = {}

    def go_fetch_sys_attrs() -> Generator[list[identifiers.SysId], None, None]:
        for page in search.fetch_sys_id_labels(container_type)(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
        ):
            sys_ids = []
            for item in page.items:
                sys_id_label_mapping[item.sys_id] = item.label
                sys_ids.append(item.sys_id)
            yield sys_ids

    output = concurrency.generate_concurrently(
        items=go_fetch_sys_attrs(),
        executor=executor,
        downstream=lambda sys_ids: _components.fetch_attribute_definitions_split(
            client=client,
            project_identifier=project_identifier,
            attribute_filter=attributes,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
            sys_ids=sys_ids,
            downstream=lambda sys_ids_split, definitions_page: concurrency.generate_concurrently(
                items=split.split_series_attributes(
                    items=(
                        identifiers.RunAttributeDefinition(
                            run_identifier=identifiers.RunIdentifier(project_identifier, sys_id),
                            attribute_definition=definition,
                        )
                        for sys_id in sys_ids_split
                        for definition in definitions_page.items
                    ),
                ),
                executor=executor,
                downstream=lambda run_attribute_definitions_split: concurrency.return_value(
                    series.fetch_series_values(
                        client=client,
                        run_attribute_definitions=run_attribute_definitions_split,
                        include_inherited=lineage_to_the_root,
                        container_type=container_type,
                        step_range=step_range,
                        tail_limit=tail_limit,
                    ),
                ),
            ),
        ),
    )

    results: Generator[
        dict[identifiers.RunAttributeDefinition, list[SeriesValue]],
        None,
        None,
    ] = concurrency.gather_results(output)

    series_data: dict[identifiers.RunAttributeDefinition, list[series.SeriesValue]] = {}
    for result in results:
        for run_attribute_definition, series_values in result.items():
            series_data.setdefault(run_attribute_definition, []).extend(series_values)

    return series_data, sys_id_label_mapping



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AttributeWarning(UserWarning):
    """Warning for attribute issues."""


class ExperimentalWarning(UserWarning):
    """Warning for use of experimental API elements."""


class Http429Warning(UserWarning):
    """Warning for retryable HTTP 429 responses (rate limiting)."""


class Http503Warning(UserWarning):
    """Warning for retryable HTTP 503 responses (service unavailable)."""


class Http5xxWarning(UserWarning):
    """Warning for retryable HTTP 5xx responses (server errors)."""


class HttpOtherWarning(UserWarning):
    """Warning for other retryable HTTP issues."""



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import (
    Generator,
    Literal,
    Optional,
)

import pandas as pd

from ...exceptions import NeptuneUserError
from .. import client as _client
from .. import context as _context
from .. import identifiers
from ..composition import attribute_components as _components
from ..composition import (
    concurrency,
    type_inference,
    validation,
)
from ..filters import (
    _Attribute,
    _BaseAttributeFilter,
    _Filter,
)
from ..identifiers import ProjectIdentifier
from ..output_format import (
    TableRow,
    create_runs_table,
    create_runs_table_multiproject,
)
from ..retrieval import attribute_values as att_vals
from ..retrieval import (
    global_search,
    search,
    util,
)

__all__ = ("fetch_table", "fetch_table_global")


def fetch_table(
    *,
    project_identifier: ProjectIdentifier,
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    sort_by: _Attribute,
    sort_direction: Literal["asc", "desc"],
    limit: Optional[int],
    type_suffix_in_column_names: bool,
    context: Optional[_context.Context] = None,
    container_type: search.ContainerType,
) -> pd.DataFrame:
    validation.validate_limit(limit)
    _sort_direction = validation.validate_sort_direction(sort_direction)

    valid_context = _context.validate_context(context or _context.get_context())
    client = _client.get_client(context=valid_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):

        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        filter_ = inference_result.get_result_or_raise()
        inference_result.emit_warnings()

        sort_by_inference_result = type_inference.infer_attribute_types_in_sort_by(
            client=client,
            project_identifier=project_identifier,
            sort_by=sort_by,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        sort_by = sort_by_inference_result.result
        sort_by_inference_result.emit_warnings()

        sys_id_label_mapping: dict[identifiers.SysId, str] = {}
        result_by_id: dict[identifiers.SysId, list[att_vals.AttributeValue]] = {}

        def go_fetch_sys_attrs() -> Generator[list[identifiers.SysId], None, None]:
            for page in search.fetch_sys_id_labels(container_type)(
                client=client,
                project_identifier=project_identifier,
                filter_=filter_,
                sort_by=sort_by,
                sort_direction=_sort_direction,
                limit=limit,
            ):
                sys_ids = []
                for item in page.items:
                    result_by_id[item.sys_id] = []  # I assume that dict preserves the order set here
                    sys_id_label_mapping[item.sys_id] = item.label
                    sys_ids.append(item.sys_id)
                yield sys_ids

        output = concurrency.generate_concurrently(
            items=go_fetch_sys_attrs(),
            executor=executor,
            downstream=lambda sys_ids: _components.fetch_attribute_definitions_split(
                client=client,
                project_identifier=project_identifier,
                attribute_filter=attributes,
                executor=executor,
                fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                sys_ids=sys_ids,
                downstream=lambda sys_ids_split, definitions_page: _components.fetch_attribute_values_split(
                    client=client,
                    project_identifier=project_identifier,
                    executor=executor,
                    sys_ids=sys_ids_split,
                    attribute_definitions=definitions_page.items,
                    downstream=concurrency.return_value,
                ),
            ),
        )

        attribute_value_pages: Generator[util.Page[att_vals.AttributeValue], None, None] = concurrency.gather_results(
            output
        )
        for attribute_values_page in attribute_value_pages:
            for attribute_value in attribute_values_page.items:
                result_by_id[attribute_value.run_identifier.sys_id].append(attribute_value)

    return create_runs_table(
        table_rows=[
            TableRow(values=values, label=sys_id_label_mapping[sys_id], project_identifier=project_identifier)
            for sys_id, values in result_by_id.items()
        ],
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=container_type,
    )


def fetch_table_global(
    *,
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    sort_by: _Attribute,
    sort_direction: Literal["asc", "desc"],
    limit: Optional[int],
    type_suffix_in_column_names: bool,
    context: Optional[_context.Context] = None,
    container_type: search.ContainerType,
) -> pd.DataFrame:
    validation.validate_limit(limit)
    _sort_direction = validation.validate_sort_direction(sort_direction)

    if sort_by.type is None:
        raise NeptuneUserError(
            "Cannot resolve sort_by attribute type. Please pass filters.Attribute with an explicit type."
        )

    type_inference.ensure_attribute_types_provided_in_filter(filter_)

    valid_context = _context.validate_context(context or _context.get_context())
    client = _client.get_client(context=valid_context)

    # Order is preserved here
    table_rows: dict[identifiers.RunIdentifier, TableRow] = {}

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):

        def go_fetch_entries() -> Generator[tuple[identifiers.ProjectIdentifier, list[identifiers.SysId]], None, None]:
            for page in global_search.fetch_global_entries(
                client=client,
                filter_=filter_,
                sort_by=sort_by,
                sort_direction=_sort_direction,
                limit=limit,
                container_type=container_type,
            ):
                entries_by_project: dict[identifiers.ProjectIdentifier, list[identifiers.SysId]] = defaultdict(list)
                for entry in page.items:
                    entries_by_project[entry.project_identifier].append(entry.sys_id)
                    table_rows[identifiers.RunIdentifier(entry.project_identifier, entry.sys_id)] = TableRow(
                        values=[],
                        label=entry.label,
                        project_identifier=entry.project_identifier,
                    )

                for project_identifier, sys_ids in entries_by_project.items():
                    yield project_identifier, sys_ids

        def _fetch_attribute_values(
            project_identifier: identifiers.ProjectIdentifier,
            sys_ids: list[identifiers.SysId],
        ) -> concurrency.OUT:
            return _components.fetch_attribute_values_by_filter_split(
                client=client,
                project_identifier=project_identifier,
                executor=executor,
                fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                sys_ids=sys_ids,
                attribute_filter=attributes,
                downstream=concurrency.return_value,
            )

        output = concurrency.generate_concurrently(
            items=go_fetch_entries(),
            executor=executor,
            downstream=lambda item: _fetch_attribute_values(project_identifier=item[0], sys_ids=item[1]),
        )

        attribute_value_pages: Generator[util.Page[att_vals.AttributeValue], None, None] = concurrency.gather_results(
            output
        )

        for attribute_values_page in attribute_value_pages:
            for attribute_value in attribute_values_page.items:
                table_rows[attribute_value.run_identifier].values.append(attribute_value)

    return create_runs_table_multiproject(
        table_rows=list(table_rows.values()),
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=container_type,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pathlib
from typing import (
    Iterable,
    Optional,
    Union,
)

import pandas as pd

from neptune_query import filters
from neptune_query.exceptions import NeptuneProjectNotProvided
from neptune_query.internal import filters as _filters
from neptune_query.internal.context import get_context
from neptune_query.internal.identifiers import ProjectIdentifier
from neptune_query.types import File

# This module contains utility functions to resolve parameters to public functions from neptune_query
# and translates them to internal objects like _Filter and _Attribute that are used in the internal API.


def resolve_experiments_filter(
    experiments: Optional[Union[str, list[str], filters.Filter]],
) -> Optional[_filters._Filter]:
    if experiments is None:
        return None
    if isinstance(experiments, str):
        return filters.Filter.name(experiments)._to_internal()
    if experiments == []:
        raise ValueError(
            "Invalid type for `experiments` filter. Expected str, non-empty list of str, or Filter object, but got "
            "an empty list"
        )
    if isinstance(experiments, list):
        # Right now, there's no Filter.any() in the public API, so we use the internal _Filter.any()
        # (The reason for not having Filter.any() is that we cannot express TRUE/FALSE in NQL
        # which would be the equivalent of Filter.any with an empty list)
        return _filters._Filter.any(
            [
                _filters._Filter.eq(filters.Attribute("sys/name", type="string")._to_internal(), exp_name)
                for exp_name in experiments
            ]
        )
    if isinstance(experiments, filters.Filter):
        return experiments._to_internal()
    raise ValueError(
        "Invalid type for `experiments` filter. Expected str, non-empty list of str, or Filter object, but got "
        f"{type(experiments)}."
    )


def resolve_attributes_filter(
    attributes: Optional[Union[str, list[str], filters.AttributeFilter]],
) -> _filters._BaseAttributeFilter:
    if attributes is None:
        return filters.AttributeFilter()._to_internal()
    if isinstance(attributes, str):
        return filters.AttributeFilter(name=attributes)._to_internal()
    if isinstance(attributes, list):
        return filters.AttributeFilter(name=attributes)._to_internal()
    if isinstance(attributes, filters.BaseAttributeFilter):
        return attributes._to_internal()
    raise ValueError(
        "Invalid type for `attributes` filter. Expected str, non-empty list of str, or AttributeFilter object, "
        f"but got {type(attributes)}."
    )


def resolve_metrics_y(
    attributes: Optional[Union[str, list[str], filters.AttributeFilter]],
) -> _filters._BaseAttributeFilter:
    return resolve_attributes_filter(attributes)


def resolve_sort_by(sort_by: Union[str, filters.Attribute]) -> _filters._Attribute:
    if isinstance(sort_by, str):
        return filters.Attribute(sort_by)._to_internal()
    if isinstance(sort_by, filters.Attribute):
        return sort_by._to_internal()
    raise ValueError(f"Invalid type for `sort_by`. Expected str or Attribute object, but got {type(sort_by)}.")


def resolve_runs_filter(runs: Optional[Union[str, list[str], filters.Filter]]) -> Optional[_filters._Filter]:
    if isinstance(runs, str):
        return filters.Filter.matches(filters.Attribute("sys/custom_run_id", type="string"), runs)._to_internal()
    if runs == []:
        raise ValueError(
            "Invalid type for `runs` filter. Expected str, non-empty list of str, or Filter object, but got "
            "an empty list"
        )
    if isinstance(runs, list):
        return _filters._Filter.any(
            [_filters._Filter.eq(_filters._Attribute("sys/custom_run_id", type="string"), run) for run in runs]
        )
    if isinstance(runs, filters.Filter):
        return runs._to_internal()
    if runs is None:
        return None
    raise ValueError(
        f"Invalid type for `runs` filter. Expected str, non-empty list of str, or Filter object, but got {type(runs)}."
    )


def get_default_project_identifier(project: Optional[str] = None) -> ProjectIdentifier:
    """
    Pass through the project name from the argument if set, otherwise, get one from env.
    """
    if not project:
        project = get_context().project

    if not project:
        raise NeptuneProjectNotProvided()

    return ProjectIdentifier(project)


def resolve_files(
    files: Union[File, Iterable[File], pd.Series, pd.DataFrame],
) -> list[File]:
    """
    Resolves the input files to a list of File objects.
    If a DataFrame is provided, it is converted to a File.
    """
    if isinstance(files, File):
        return [files]
    elif isinstance(files, pd.Series):
        return [file for file in files if isinstance(file, File)]
    elif isinstance(files, pd.DataFrame):
        return [file for _, col in files.items() for file in col if isinstance(file, File)]
    elif isinstance(files, Iterable):
        return [file for file in files if isinstance(file, File)]
    else:
        raise ValueError(
            f"Invalid type for `files`. Expected File, DataFrame, Series, "
            f"or iterable of Files, but got {type(files)}."
        )


def resolve_destination_path(destination: Optional[Union[str, pathlib.Path]]) -> pathlib.Path:
    if destination is None:
        return pathlib.Path.cwd()
    elif isinstance(destination, pathlib.Path):
        return destination.resolve()
    else:
        return pathlib.Path(destination).resolve()



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import base64
import contextlib
import functools
import json
import logging
import secrets
from importlib.metadata import (
    PackageNotFoundError,
    version,
)
from typing import (
    Callable,
    Generator,
    Optional,
    ParamSpec,
    TypeVar,
)

from neptune_api.types import Response

from neptune_query.internal import env
from neptune_query.internal.composition import concurrency

# This flag is used to control whether query metadata should be added to the request headers
ADD_QUERY_METADATA = True

logger = logging.getLogger(__name__)


class QueryMetadata:
    def __init__(
        self,
        api_function: str,
        client_version: str,
        nq_query_id: str,
        user_data: str | None = None,
    ) -> None:

        # longest observed: "fetch_experiments_table_global" - 30 characters
        self.api_function = api_function[:32]

        # longest observed: "nq/1.8.0b1.post8+cea1d73" - 24 characters
        self.client_version = client_version[:24]

        # nq_query_id should be exactly 8 characters
        self.nq_query_id = nq_query_id[:8]

        # user_data may contribute at most 80 characters when JSON-encoded
        self.user_data: str | dict | None = _process_user_data(user_data)

    def to_json(self) -> str:
        return json.dumps(
            {
                "fn": self.api_function,
                "v": self.client_version,
                "qid": self.nq_query_id,
                "ud": self.user_data,
            }
        )


@contextlib.contextmanager
def use_query_metadata(api_function: str) -> Generator[None, None, None]:
    user_data = env.NEPTUNE_QUERY_METADATA.get()

    query_metadata = QueryMetadata(
        api_function=api_function,
        client_version=_get_client_version(),
        nq_query_id=_generate_nq_query_id(),
        user_data=user_data if user_data else None,
    )
    with concurrency.use_thread_local({"query_metadata": query_metadata}):
        yield


T = ParamSpec("T")
R = TypeVar("R")


def with_neptune_client_metadata(func: Callable[T, Response[R]]) -> Callable[T, Response[R]]:
    @functools.wraps(func)
    def wrapper(*args: T.args, **kwargs: T.kwargs) -> Response[R]:
        query_metadata: Optional[QueryMetadata] = concurrency.get_thread_local(
            "query_metadata", expected_type=QueryMetadata
        )
        if ADD_QUERY_METADATA and query_metadata:
            kwargs["x_neptune_client_metadata"] = query_metadata.to_json()
        return func(*args, **kwargs)

    return wrapper


@functools.cache
def _get_client_version() -> str:
    try:
        package_version = version("neptune-query")
    except PackageNotFoundError:
        package_version = "unknown"
    return f"nq/{package_version}"


@functools.cache
def _process_user_data(user_data: str | None) -> str | dict | None:
    if not user_data:
        return None

    try:
        user_data = json.loads(user_data)
        user_data_is_proper_json = True
    except json.JSONDecodeError:
        user_data_is_proper_json = False

    if len(json.dumps(user_data)) > 82:
        user_data = "NEPTUNE_QUERY_METADATA too long"
        if user_data_is_proper_json:
            logger.debug("User data in NEPTUNE_QUERY_METADATA env too long. Skipping user data in query metadata.")
        else:
            logger.debug(
                "User data in NEPTUNE_QUERY_METADATA env too long (and not JSON-encoded). "
                "Skipping user data in query metadata."
            )

    return user_data


def _generate_nq_query_id() -> str:
    # Generate a random string of exactly 8 URL-safe base64 characters (a-z, A-Z, 0-9, -, _):
    nq_query_id_bytes = secrets.token_bytes(6)
    nq_query_id = base64.urlsafe_b64encode(nq_query_id_bytes).decode("ascii")
    return nq_query_id



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from typing import (
    Generator,
    Optional,
)

import pandas as pd

from ... import types
from .. import client as _client
from ..composition import (
    concurrency,
    validation,
)
from ..context import (
    Context,
    get_context,
    validate_context,
)
from ..output_format import create_files_dataframe
from ..retrieval import files as _files
from ..retrieval.search import ContainerType
from ..retrieval.split import split_files


def download_files(
    *,
    files: list[types.File],
    destination: pathlib.Path,
    container_type: ContainerType,
    context: Optional[Context] = None,
) -> pd.DataFrame:
    validation.ensure_write_access(destination)
    valid_context = validate_context(context or get_context())
    client = _client.get_client(context=valid_context)

    with concurrency.create_thread_pool_executor() as executor:

        def generate_signed_files() -> Generator[tuple[types.File, _files.SignedFile], None, None]:
            for file_group in split_files(files):
                signed_files = _files.fetch_signed_urls(client=client, files=file_group)
                yield from zip(file_group, signed_files)

        output = concurrency.generate_concurrently(
            items=generate_signed_files(),
            executor=executor,
            downstream=lambda file_tuple: concurrency.return_value(
                (
                    file_tuple[0],
                    _files.download_file_complete(
                        client=client,
                        signed_file=file_tuple[1],
                        target_path=_files.create_target_path(
                            destination=destination,
                            file=file_tuple[0],
                        ),
                    ),
                )
            ),
        )

        results: Generator[
            tuple[types.File, Optional[pathlib.Path]],
            None,
            None,
        ] = concurrency.gather_results(output)

        file_paths: dict[types.File, Optional[pathlib.Path]] = {}
        for file, path in results:
            file_paths[file] = path

        return create_files_dataframe(
            file_paths,
            container_type=container_type,
        )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import (
    Generator,
    Iterable,
    TypeVar,
)

from .. import (
    env,
    identifiers,
)
from ..identifiers import RunAttributeDefinition

_UUID_SIZE = 50


T = TypeVar("T")


def _attribute_definition_size(attr: identifiers.AttributeDefinition) -> int:
    return _attribute_name_size(attr.name)


def _attribute_name_size(attr_name: str) -> int:
    return len(attr_name.encode("utf-8"))


def _sys_id_size() -> int:
    return _UUID_SIZE


def split_sys_ids(
    sys_ids: list[identifiers.SysId],
) -> Generator[list[identifiers.SysId]]:
    """
    Splits a sequence of sys ids into batches of size at most `NEPTUNE_QUERY_MAX_REQUEST_SIZE`.
    Use before fetching attribute definitions.
    """
    query_size_limit = env.NEPTUNE_QUERY_MAX_REQUEST_SIZE.get()
    identifier_num_limit = max(query_size_limit // _sys_id_size(), 1)

    identifier_num = len(sys_ids)
    batch_num = _ceil_div(identifier_num, identifier_num_limit)

    if batch_num == 1:
        yield sys_ids
    elif batch_num > 1:
        batch_size = _ceil_div(identifier_num, batch_num)
        for i in range(0, identifier_num, batch_size):
            yield sys_ids[i : i + batch_size]


def split_attribute_names(attribute_names: list[str]) -> list[list[str]]:
    query_size_limit = env.NEPTUNE_QUERY_MAX_ATTRIBUTE_FILTER_SIZE.get()

    attribute_batches = []
    current_batch: list[str] = []
    current_batch_size = 0
    for attr in attribute_names:
        attr_size = _attribute_name_size(attr)
        if current_batch and current_batch_size + attr_size > query_size_limit:
            attribute_batches.append(current_batch)
            current_batch = []
            current_batch_size = 0
        current_batch.append(attr)
        current_batch_size += attr_size

    if current_batch:
        attribute_batches.append(current_batch)

    return attribute_batches


def split_sys_ids_attributes(
    sys_ids: list[identifiers.SysId],
    attribute_definitions: list[identifiers.AttributeDefinition],
) -> Generator[tuple[list[identifiers.SysId], list[identifiers.AttributeDefinition]]]:
    """
    Splits a pair of sys ids and attribute_definitions into batches that:
    When their length is added it is of size at most `NEPTUNE_QUERY_MAX_REQUEST_SIZE`.
    When their item count is multiplied, it is at most `NEPTUNE_QUERY_ATTRIBUTE_VALUES_BATCH_SIZE`.

    It's intended for use before fetching attribute values and assumes that the sys_ids and attribute_definitions
    will be sent to the server in a single request and the response will contain data for their cartesian product.
    """
    query_size_limit = env.NEPTUNE_QUERY_MAX_REQUEST_SIZE.get()
    attribute_values_batch_size = env.NEPTUNE_QUERY_ATTRIBUTE_VALUES_BATCH_SIZE.get()

    if not attribute_definitions:
        return

    attribute_batches = _split_attribute_definitions(
        attribute_definitions,
        query_size_limit=query_size_limit - _sys_id_size(),  # ensure at least one sys_id can fit later
        attribute_values_batch_size=attribute_values_batch_size,
    )
    max_attribute_batch_size = max(
        sum(_attribute_definition_size(attr) for attr in batch) for batch in attribute_batches
    )
    max_attribute_batch_len = max(len(batch) for batch in attribute_batches)

    sys_id_batch: list[identifiers.SysId] = []
    total_batch_size = max_attribute_batch_size
    for experiment in sys_ids:
        if sys_id_batch and (
            (len(sys_id_batch) + 1) * max_attribute_batch_len > attribute_values_batch_size
            or total_batch_size + _sys_id_size() > query_size_limit
        ):
            for attribute_batch in attribute_batches:
                yield sys_id_batch, attribute_batch
            sys_id_batch = []
            total_batch_size = max_attribute_batch_size
        sys_id_batch.append(experiment)
        total_batch_size += _sys_id_size()
    if sys_id_batch:
        for attribute_batch in attribute_batches:
            yield sys_id_batch, attribute_batch


def _split_attribute_definitions(
    attribute_definitions: list[identifiers.AttributeDefinition],
    query_size_limit: int,
    attribute_values_batch_size: int,
) -> list[list[identifiers.AttributeDefinition]]:

    attribute_batches = []
    current_batch: list[identifiers.AttributeDefinition] = []
    current_batch_size = 0
    for attr in attribute_definitions:
        attr_size = _attribute_definition_size(attr)
        if current_batch and (
            len(current_batch) >= attribute_values_batch_size or current_batch_size + attr_size > query_size_limit
        ):
            attribute_batches.append(current_batch)
            current_batch = []
            current_batch_size = 0
        current_batch.append(attr)
        current_batch_size += attr_size

    if current_batch:
        attribute_batches.append(current_batch)

    return attribute_batches


def split_series_attributes(items: Iterable[RunAttributeDefinition]) -> Generator[list[RunAttributeDefinition]]:
    """
    Splits a list of classes containing an attribute_definition into batches so that:
    When the lengths of attribute paths are added, the total length is at most `NEPTUNE_QUERY_MAX_REQUEST_SIZE`.
    Item count is at most `NEPTUNE_QUERY_SERIES_BATCH_SIZE`.

    Intended for use before fetching (string, float) series.
    """
    query_size_limit = env.NEPTUNE_QUERY_MAX_REQUEST_SIZE.get()
    batch_size_limit = env.NEPTUNE_QUERY_SERIES_BATCH_SIZE.get()

    if not items:
        return

    batch: list[RunAttributeDefinition] = []
    batch_size = 0
    for item in items:
        attr_size = _attribute_name_size(item.attribute_definition.name)
        if batch and (len(batch) >= batch_size_limit or batch_size + attr_size > query_size_limit):
            yield batch
            batch = []
            batch_size = 0
        batch.append(item)
        batch_size += attr_size

    if batch:
        yield batch


def _ceil_div(a: int, b: int) -> int:
    return (a + b - 1) // b


def split_files(items: list[T]) -> Generator[list[T], None, None]:
    batch_size = env.NEPTUNE_QUERY_FILES_BATCH_SIZE.get()

    for i in range(0, len(items), batch_size):
        yield items[i : i + batch_size]



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import functools as ft
from dataclasses import dataclass
from typing import (
    Generator,
    Literal,
    Optional,
)

import attrs
from neptune_api.api.retrieval import search_global_leaderboard_entries_proto
from neptune_api.client import AuthenticatedClient
from neptune_api.models import (
    AttributeTypeDTO,
    GlobalSearchParamsDTO,
    NqlQueryParamsDTO,
    QueryLeaderboardParamsFieldDTO,
    QueryLeaderboardParamsFieldDTOAggregationMode,
    QueryLeaderboardParamsPaginationDTO,
    QueryLeaderboardParamsSortingParamsDTO,
    QueryLeaderboardParamsSortingParamsDTODir,
)
from neptune_api.proto.neptune_pb.api.v1.model.leaderboard_entries_pb2 import (
    ProtoAttributesDTO,
    ProtoLeaderboardEntriesSearchResultDTO,
)
from neptune_api.types import UNSET

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from .. import env
from ..filters import (
    _Attribute,
    _Filter,
)
from ..identifiers import (
    CustomRunId,
    ProjectIdentifier,
    SysId,
    SysName,
)
from ..logger import get_logger
from ..retrieval import (
    retry,
    util,
)
from ..retrieval.attribute_types import map_attribute_type_python_to_backend
from .search import ContainerType

logger = get_logger()

__all__ = ("GlobalRunSearchEntry", "fetch_global_entries")


@dataclass(frozen=True)
class GlobalRunSearchEntry:
    sys_id: SysId
    sys_name: Optional[SysName]
    sys_custom_run_id: Optional[CustomRunId]
    project_identifier: ProjectIdentifier
    container_type: ContainerType

    @property
    def label(self) -> str:
        return str(self.sys_name) if self.container_type == ContainerType.EXPERIMENT else str(self.sys_custom_run_id)

    def __post_init__(self) -> None:
        if self.sys_id is None:
            raise ValueError(f"sys_id cannot be None in a global search entry ({self})")
        if self.container_type == ContainerType.RUN and self.sys_custom_run_id is None:
            raise ValueError(f"sys/custom_run_id missing for a global search entry representing run ({self})")
        if self.container_type == ContainerType.EXPERIMENT and self.sys_name is None:
            raise ValueError(f"sys/name missing for a global search entry representing experiment ({self})")


def fetch_global_entries(
    *,
    client: AuthenticatedClient,
    filter_: Optional[_Filter],
    sort_by: _Attribute,
    sort_direction: Literal["asc", "desc"],
    limit: Optional[int],
    container_type: ContainerType,
) -> Generator[util.Page[GlobalRunSearchEntry], None, None]:
    batch_size = env.NEPTUNE_QUERY_SYS_ATTRS_BATCH_SIZE.get()

    yield from util.fetch_pages(
        client=client,
        fetch_page=_fetch_entries_page,
        process_page=ft.partial(_process_entries_page, container_type=container_type),
        make_new_page_params=ft.partial(_make_next_page_params, limit=limit, batch_size=batch_size),
        initial_params=GlobalSearchParamsDTO(
            experiment_leader=container_type == ContainerType.EXPERIMENT,
            pagination=QueryLeaderboardParamsPaginationDTO(
                limit=min(batch_size, limit) if limit is not None else batch_size,
                offset=0,
            ),
            query=NqlQueryParamsDTO(query=filter_.to_query()) if filter_ is not None else UNSET,
            sorting=QueryLeaderboardParamsSortingParamsDTO(
                sort_by=QueryLeaderboardParamsFieldDTO(
                    name=sort_by.name,
                    type=AttributeTypeDTO(
                        map_attribute_type_python_to_backend(sort_by.type) if sort_by.type is not None else "string"
                    ),
                    aggregation_mode=QueryLeaderboardParamsFieldDTOAggregationMode(sort_by.aggregation)
                    if sort_by.aggregation is not None
                    else UNSET,
                ),
                dir_=QueryLeaderboardParamsSortingParamsDTODir.ASCENDING
                if sort_direction == "asc"
                else QueryLeaderboardParamsSortingParamsDTODir.DESCENDING,
            ),
        ),
    )


def _fetch_entries_page(
    client: AuthenticatedClient,
    params: GlobalSearchParamsDTO,
) -> ProtoLeaderboardEntriesSearchResultDTO:
    logger.debug(f"Calling search_l_global_entries_proto with params: {params}")

    call_api = retry.handle_errors_default(
        with_neptune_client_metadata(search_global_leaderboard_entries_proto.sync_detailed)
    )
    response = call_api(client=client, body=params)

    logger.debug(
        f"search_l_global_entries_proto response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )
    return ProtoLeaderboardEntriesSearchResultDTO.FromString(response.content)


def _make_next_page_params(
    current_params: GlobalSearchParamsDTO,
    data: Optional[ProtoLeaderboardEntriesSearchResultDTO],
    *,
    limit: Optional[int],
    batch_size: int,
) -> Optional[GlobalSearchParamsDTO]:
    if data is None:
        return current_params

    # If the server returned fewer entries than we process in a single request, we've reached the end.
    if len(data.entries) < batch_size:
        return None

    retrieved_so_far = current_params.pagination.offset + batch_size
    if limit is not None and retrieved_so_far >= limit:
        return None

    return attrs.evolve(
        current_params,
        pagination=attrs.evolve(
            current_params.pagination,
            limit=min(limit - retrieved_so_far, batch_size) if limit is not None else batch_size,
            offset=retrieved_so_far,
        ),
    )


def _process_entries_page(
    data: ProtoLeaderboardEntriesSearchResultDTO,
    *,
    container_type: ContainerType,
) -> util.Page[GlobalRunSearchEntry]:
    def to_global_search_entry(entry: ProtoAttributesDTO) -> GlobalRunSearchEntry:
        attributes: dict[str, str] = {
            attribute.name: attribute.string_properties.value
            for attribute in entry.attributes
            if attribute.HasField("string_properties")
        }

        return GlobalRunSearchEntry(
            sys_id=SysId(attributes["sys/id"]),
            sys_name=SysName(attributes["sys/name"]) if attributes.get("sys/name") is not None else None,
            sys_custom_run_id=(
                CustomRunId(attributes["sys/custom_run_id"])
                if attributes.get("sys/custom_run_id") is not None
                else None
            ),
            project_identifier=ProjectIdentifier(f"{entry.organization_name}/{entry.project_name}"),
            container_type=container_type,
        )

    return util.Page(items=[to_global_search_entry(entry) for entry in data.entries])



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "fetch_experiments_table_global",
    "fetch_runs_table_global",
]

from typing import (
    Literal,
    Optional,
    Union,
)

import pandas as _pandas

from neptune_query import filters
from neptune_query._internal import (
    resolve_attributes_filter,
    resolve_experiments_filter,
    resolve_runs_filter,
    resolve_sort_by,
)
from neptune_query.internal.composition import fetch_table as _fetch_table
from neptune_query.internal.experimental import experimental as experimental_api
from neptune_query.internal.query_metadata_context import use_query_metadata
from neptune_query.internal.retrieval import search as _search


@use_query_metadata(api_function="fetch_experiments_table_global")
@experimental_api
def fetch_experiments_table_global(
    *,
    experiments: Optional[Union[str, list[str], filters.Filter]] = None,
    attributes: Union[str, list[str], filters.AttributeFilter] = [],
    sort_by: filters.Attribute = filters.Attribute("sys/creation_time", type="datetime"),
    sort_direction: Literal["asc", "desc"] = "desc",
    limit: Optional[int] = None,
    type_suffix_in_column_names: bool = False,
) -> _pandas.DataFrame:
    """Fetches a table of experiment metadata, with runs as rows and attributes as columns.

    To narrow the results, define filters for experiments to search or attributes to include.

    Returns a DataFrame similar to the runs table in the web app.
    For series attributes, the last logged value is returned.

    Args:
        experiments: Filter specifying which experiments to include.
            If a string is provided, it's treated as a regex pattern that the names must match.
            If a list of strings is provided, it's treated as exact experiment names to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.
            The filter must use only selected attributes from the sys/ and env/ namespaces. Attribute types
            must be provided explicitly in the filter (use filters.Attribute with a type).
            Ask your administrator for a list of supported attributes.
        attributes: Filter specifying which attributes to include.
            If a string is provided, it's treated as a regex pattern that the attribute names must match.
            If a list of strings is provided, it's treated as exact attribute names to match.
            To provide a more complex condition, pass an AttributeFilter object.
        sort_by: Attribute to sort the table by. If specified, needs to specify the attribute type.
        sort_direction: The direction to sort columns by: `"desc"` (default) or `"asc"`.
        limit: Maximum number of experiments to return. By default, all experiments are included.
        type_suffix_in_column_names: If True, columns of the returned DataFrame
            are suffixed with ":<type>", e.g. "attribute1:float_series", "attribute1:string".
            If False (default), the method throws an exception if there are multiple types under one path.

    Example:
        Fetch attributes matching `loss` or `configs` from two specific experiments:
        ```
        import neptune_query.experimental as nq_experimental


        nq_experimental.fetch_experiments_table_global(
            experiments=["seagull-week1", "seagull-week2"],
            attributes=r"loss | configs",
        )
        ```
    """
    experiments_filter = resolve_experiments_filter(experiments)
    attributes_filter = resolve_attributes_filter(attributes)
    resolved_sort_by = resolve_sort_by(sort_by)

    return _fetch_table.fetch_table_global(
        filter_=experiments_filter,
        attributes=attributes_filter,
        sort_by=resolved_sort_by,
        sort_direction=sort_direction,
        limit=limit,
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=_search.ContainerType.EXPERIMENT,
    )


@use_query_metadata(api_function="fetch_runs_table_global")
@experimental_api
def fetch_runs_table_global(
    *,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
    attributes: Union[str, list[str], filters.AttributeFilter] = [],
    sort_by: filters.Attribute = filters.Attribute("sys/creation_time", type="datetime"),
    sort_direction: Literal["asc", "desc"] = "desc",
    limit: Optional[int] = None,
    type_suffix_in_column_names: bool = False,
) -> _pandas.DataFrame:
    """Fetches a table of run metadata, with runs as rows and attributes as columns.

    To narrow the results, define filters for runs to search or attributes to include.

    Returns a DataFrame similar to the runs table in the web app.
    For series attributes, the last logged value is returned.

    Args:
        runs: Filter specifying which runs to include.
            If a string is provided, it's treated as a regex pattern that the run IDs must match.
            If a list of strings is provided, it's treated as exact run IDs to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.
            The filter must use only attributes from the sys/ and env/ namespaces. Attribute types
            must be provided explicitly in the filter (use filters.Attribute with a type).
        attributes: Filter specifying which attributes to include.
            If a string is provided, it's treated as a regex pattern that the attribute names must match.
            If a list of strings is provided, it's treated as exact attribute names to match.
            To provide a more complex condition, pass an AttributeFilter object.
        sort_by: Attribute to sort the table by. If specified, needs to specify the attribute type.
        sort_direction: The direction to sort columns by: `"desc"` (default) or `"asc"`.
        limit: Maximum number of runs to return. By default, all runs are included.
        type_suffix_in_column_names: If True, columns of the returned DataFrame
            are suffixed with ":<type>", e.g. "attribute1:float_series", "attribute1:string".
            If False (default), the method throws an exception if there are multiple types under one path.

    Example:
        Fetch constituent runs of an experiment, with attributes matching `loss` or `configs` as columns:
        ```
        import neptune_query.experimental as nq_experimental
        from neptune_query.filters import Filter


        nq_experimental.fetch_runs_table_global(
            runs=Filter.eq("sys/name", "exp-week9"),
            attributes=r"loss | configs",
        )
        ```
    """
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)
    resolved_sort_by = resolve_sort_by(sort_by)

    return _fetch_table.fetch_table_global(
        filter_=runs_filter,
        attributes=attributes_filter,
        sort_by=resolved_sort_by,
        sort_direction=sort_direction,
        limit=limit,
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=_search.ContainerType.RUN,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import warnings
from datetime import (
    datetime,
    timedelta,
)
from typing import Type

from neptune_query.warnings import (
    ExperimentalWarning,
    Http5xxWarning,
    Http429Warning,
    HttpOtherWarning,
)

# Types of warnings that should be emitted only once per unique message
WARNING_TYPES_EMITTED_ONCE_PER_MESSAGE = (ExperimentalWarning,)

# Types of warnings that should be emitted only once per a certain time period
WARNING_TYPES_EMITTED_ONCE_PER_WHILE = (Http429Warning, Http5xxWarning, HttpOtherWarning)
WARNING_SUPPRESSION_DURATION = timedelta(seconds=20)

# registry of warnings that were already emitted with the (type, message) tuple
_silence_warnings_msg: set[tuple[Type[Warning], str]] = set()

# registry of warning types that were already emitted with the time they should be silenced until
_silence_warnings_until: dict[Type[Warning], datetime] = {}


def format_warning(warning: Warning) -> Warning:
    # check if stderr is a terminal:
    if sys.stderr.isatty():
        orange_bold = "\033[1;38;2;255;165;0m"
        end = "\033[0m"
        msg = f"{orange_bold}{str(warning)}{end}"
        return type(warning)(msg)
    else:
        return warning


def throttled_warn(warning: Warning, stacklevel: int = 3) -> None:
    if isinstance(warning, WARNING_TYPES_EMITTED_ONCE_PER_MESSAGE):
        key = (type(warning), str(warning))
        if key in _silence_warnings_msg:
            return
        _silence_warnings_msg.add(key)

    if isinstance(warning, WARNING_TYPES_EMITTED_ONCE_PER_WHILE):
        warning_type = type(warning)
        now = datetime.now()
        if warning_type in _silence_warnings_until and _silence_warnings_until[warning_type] > now:
            return
        _silence_warnings_until[warning_type] = now + WARNING_SUPPRESSION_DURATION

    warnings.warn(format_warning(warning), stacklevel=stacklevel)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools as ft
from typing import (
    Any,
    Generator,
    Iterable,
    Optional,
)

from neptune_api.api.retrieval import query_attribute_definitions_within_project
from neptune_api.client import AuthenticatedClient
from neptune_api.models import (
    QueryAttributeDefinitionsBodyDTO,
    QueryAttributeDefinitionsResultDTO,
)

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from .. import filters  # noqa: E402
from .. import (  # noqa: E402
    env,
    identifiers,
)
from ..logger import get_logger
from ..retrieval import attribute_types as types  # noqa: E402
from ..retrieval import util  # noqa: E402
from ..retrieval import retry
from .attribute_filter import transform_attribute_filter_into_params

logger = get_logger()


def fetch_attribute_definitions_single_filter(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Optional[Iterable[identifiers.RunIdentifier]],
    attribute_filter: filters._AttributeFilter,
    batch_size: int = env.NEPTUNE_QUERY_ATTRIBUTE_DEFINITIONS_BATCH_SIZE.get(),
) -> Generator[util.Page[identifiers.AttributeDefinition], None, None]:
    params: dict[str, Any] = {
        "projectIdentifiers": list(project_identifiers),
        "attributeNameFilter": dict(),
        "nextPage": {"limit": batch_size},
    }

    if run_identifiers is not None:
        params["experimentIdsFilter"] = [str(e) for e in run_identifiers]

    attribute_filter_params = transform_attribute_filter_into_params(attribute_filter)
    params.update(attribute_filter_params)

    return util.fetch_pages(
        client=client,
        fetch_page=_fetch_attribute_definitions_page,
        process_page=_process_attribute_definitions_page,
        make_new_page_params=ft.partial(_make_new_attribute_definitions_page_params, batch_size=batch_size),
        initial_params=params,
    )


def _fetch_attribute_definitions_page(
    client: AuthenticatedClient,
    params: dict[str, Any],
) -> QueryAttributeDefinitionsResultDTO:
    logger.debug(f"Calling query_attribute_definitions_within_project with params: {params}")

    body = QueryAttributeDefinitionsBodyDTO.from_dict(params)
    call_api = retry.handle_errors_default(
        with_neptune_client_metadata(query_attribute_definitions_within_project.sync_detailed)
    )
    response = call_api(client=client, body=body)

    logger.debug(
        f"query_attribute_definitions_within_project response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )

    return response.parsed


def _process_attribute_definitions_page(
    data: QueryAttributeDefinitionsResultDTO,
) -> util.Page[identifiers.AttributeDefinition]:
    items = []
    for entry in data.entries:
        item = identifiers.AttributeDefinition(
            name=entry.name,
            type=types.map_attribute_type_backend_to_python(str(entry.type)),
        )
        items.append(item)
    return util.Page(items=items)


def _make_new_attribute_definitions_page_params(
    params: dict[str, Any],
    data: Optional[QueryAttributeDefinitionsResultDTO],
    batch_size: int,
) -> Optional[dict[str, Any]]:
    if data is None:
        if "nextPageToken" in params["nextPage"]:
            del params["nextPage"]["nextPageToken"]
        return params

    next_page_token = data.next_page.next_page_token
    if not next_page_token or len(data.entries) < batch_size:
        return None

    params["nextPage"]["nextPageToken"] = next_page_token
    return params



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import itertools as it
import re
from typing import (
    Any,
    Iterable,
    Optional,
    Union,
)

from .. import filters  # noqa: E402
from ..retrieval import attribute_types as types  # noqa: E402
from .split import split_attribute_names


def split_attribute_filters(
    _attribute_filter: filters._BaseAttributeFilter,
) -> list[filters._AttributeFilter]:
    if isinstance(_attribute_filter, filters._AttributeFilter):
        if isinstance(_attribute_filter.name_eq, list):
            split_name_eqs = split_attribute_names(_attribute_filter.name_eq)
            return [dataclasses.replace(_attribute_filter, name_eq=name_eq_subset) for name_eq_subset in split_name_eqs]
        else:
            return [_attribute_filter]
    elif isinstance(_attribute_filter, filters._AttributeFilterAlternative):
        return list(it.chain.from_iterable(split_attribute_filters(child) for child in _attribute_filter.filters))
    else:
        raise RuntimeError(f"Unexpected filter type: {type(_attribute_filter)}")


def transform_attribute_filter_into_params(
    attribute_filter: filters._AttributeFilter,
) -> dict[str, Any]:
    params: dict[str, Any] = {
        "attributeNameFilter": {},
    }

    name_regexes = None
    if attribute_filter.name_eq is not None:
        name_regexes = _escape_name_eq(_variants_to_list(attribute_filter.name_eq))

    if attribute_filter.must_match_any is not None:
        attribute_name_filter_dtos = []
        for alternative in attribute_filter.must_match_any:
            attribute_name_filter_dto = {}
            must_match_regexes = _union_options([name_regexes, alternative.must_match_regexes])
            if must_match_regexes is not None:
                attribute_name_filter_dto["mustMatchRegexes"] = must_match_regexes
            if alternative.must_not_match_regexes is not None:
                attribute_name_filter_dto["mustNotMatchRegexes"] = alternative.must_not_match_regexes
            if attribute_name_filter_dto:
                attribute_name_filter_dtos.append(attribute_name_filter_dto)
        params["attributeNameFilter"]["mustMatchAny"] = attribute_name_filter_dtos

    elif name_regexes is not None:
        params["attributeNameFilter"]["mustMatchAny"] = [{"mustMatchRegexes": name_regexes}]

    attribute_types = _variants_to_list(attribute_filter.type_in)
    if attribute_types is not None:
        params["attributeFilter"] = [
            {"attributeType": types.map_attribute_type_python_to_backend(_type)} for _type in attribute_types
        ]

    # note: attribute_filter.aggregations is intentionally ignored

    return params


def _escape_name_eq(names: Optional[list[str]]) -> Optional[list[str]]:
    if names is None:
        return None

    escaped = [f"{re.escape(name)}" for name in names]

    if len(escaped) == 1:
        return [f"^{escaped[0]}$"]
    else:
        joined = "|".join(escaped)
        return [f"^({joined})$"]


def _variants_to_list(param: Union[str, Iterable[str], None]) -> Optional[list[str]]:
    if param is None:
        return None
    if isinstance(param, str):
        return [param]
    return list(param)


def _union_options(options: list[Optional[list[str]]]) -> Optional[list[str]]:
    result = None

    for option in options:
        if option is not None:
            if result is None:
                result = []
            result.extend(option)

    return result
